@@ -4,8 +4,7 @@ import random
 import re
 import string
 import warnings
-from math import ceil
-from operator import itemgetter
+from html import unescape
 
 import graphviz
 import lime
@@ -14,9 +13,11 @@ import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 import shap
+import spacy
 import statsmodels.api as sm
 from gensim.corpora import Dictionary
 from gensim.models import CoherenceModel, nmf
+from nltk.corpus import stopwords
 from sklearn.decomposition import NMF
 from sklearn.feature_extraction.text import (
     ENGLISH_STOP_WORDS,
@@ -37,6 +38,7 @@ from sklearn.metrics import (
 )
 from sklearn.model_selection import train_test_split
 from sklearn.tree import export_graphviz
+from spacy.cli import download
 from tabulate import tabulate
 
 from datto.CleanDataframe import CleanDataframe
@@ -44,14 +46,47 @@ from datto.CleanDataframe import CleanDataframe
 # Hide precision/recall/f1 warnings for model scorings
 warnings.filterwarnings("always")
 
+try:
+    nlp = spacy.load("en")
+except Exception:
+    download("en")
+    nlp = spacy.load("en")
+
 
 class ModelResults:
+    def jaccard_similarity(self, topic_1, topic_2):
+        """
+        Derives the Jaccard similarity of two topics
+
+        Jaccard similarity:
+        - A statistic used for comparing the similarity and diversity of sample sets
+        - J(A,B) = (A ∩ B)/(A ∪ B)
+        - Goal is low Jaccard scores for coverage of the diverse elements
+
+        Parameters
+        --------
+        topic_1: list
+        topic_2: list
+
+
+        Returns
+        --------
+        score: float
+
+        """
+
+        intersection = set(topic_1).intersection(set(topic_2))
+        union = set(topic_1).union(set(topic_2))
+        score = float(len(intersection)) / float(len(union))
+
+        return score
+
     def most_similar_texts(
         self,
         X,
         num_examples,
         text_column_name,
-        num_topics=None,
+        chosen_num_topics=None,
         chosen_stopwords=set(),
         min_df=3,
         max_df=0.1,
@@ -66,7 +101,7 @@ class ModelResults:
         X: DataFrame
         num_examples: int
         text_column_name: str
-        num_topics: int
+        chosen_num_topics: int
             Optional - if none algorithm will determine best number
         min_df: float
             Minimum number/proportion of docs that need to have the words
@@ -93,12 +128,49 @@ class ModelResults:
         X = X[X[text_column_name] != "N/A"]
         X = X[X[text_column_name] != "na"]
 
+        # Remove HTML & unicode characters
+        X[text_column_name] = X[text_column_name].apply(
+            lambda x: unescape(x)
+            .encode("ascii", errors="ignore")
+            .decode("ascii")
+            .replace("    ", " ")
+            .replace("   ", " ")
+            .replace("  ", " ")
+            .strip()
+        )
+
         all_stop_words = (
+            # Combine stopwords from all the packages
             set(ENGLISH_STOP_WORDS)
-            | set(["-PRON-"])
+            | set(stopwords.words("english"))
+            | nlp.Defaults.stop_words
             | set(string.punctuation)
-            | set([" "])
-            | chosen_stopwords
+            | set(string.ascii_lowercase)
+            | set(
+                [
+                    "-PRON-",
+                    " ",
+                    "i.e",
+                    "e.g.",
+                    "-",
+                    "--",
+                    "---",
+                    "----",
+                    "..",
+                    "...",
+                    "....",
+                    "w/",
+                    "^^",
+                    "^^^",
+                    "^^^^",
+                    "’",
+                    "~~",
+                    "~~~",
+                    "~~~~",
+                    "and/or",
+                ]
+            )
+            | set(chosen_stopwords)
         )
 
         ct = CleanDataframe()
@@ -115,18 +187,14 @@ class ModelResults:
         vocab = vectorizer.get_feature_names()
         vector_df = pd.DataFrame(vectors, columns=vocab, index=X.index)
 
-        if X.shape[0] < 20:
+        if X.shape[0] <= 75:
             return "Too few examples to categorize."
 
-        if not num_topics:
+        if not chosen_num_topics:
+            # Inspired by: https://bit.ly/3CqH2Zw
 
-            # In case 1, add 1 to get at least 2
-            # The rest are based on eyeballing numbers
-            min_topics = ceil(X.shape[0] * 0.01) + 1
-            max_topics = ceil(X.shape[0] * 0.2)
-            step = ceil((max_topics - min_topics) / 5)
-
-            topic_nums = list(np.arange(min_topics, max_topics, step))
+            # Test out several topic numbers
+            topic_nums = list(np.arange(5, 75 + 1, 5))
 
             texts = X[text_column_name].apply(ct.lematize)
 
@@ -139,36 +207,62 @@ class ModelResults:
             # Create the bag-of-words format (list of (token_id, token_count))
             corpus = [dictionary.doc2bow(text) for text in texts]
 
-            coherence_scores = []
+            NMF_models = {}
+            NMF_topics = {}
 
-            for num in topic_nums:
-                model = nmf.Nmf(
+            for i in topic_nums:
+                NMF_models[i] = nmf.Nmf(
                     corpus=corpus,
-                    num_topics=num,
                     id2word=dictionary,
+                    num_topics=i,
                     chunksize=2000,
-                    passes=5,
-                    kappa=0.1,
-                    minimum_probability=0.01,
-                    w_max_iter=300,
-                    w_stop_condition=0.0001,
-                    h_max_iter=100,
-                    h_stop_condition=0.001,
-                    eval_every=10,
-                    normalize=True,
+                    passes=20,
                     random_state=42,
                 )
 
-                cm = CoherenceModel(
-                    model=model, texts=texts, dictionary=dictionary, coherence="u_mass"
+                shown_topics = NMF_models[i].show_topics(
+                    num_topics=i, num_words=15, formatted=False
                 )
+                NMF_topics[i] = [
+                    [word[0] for word in topic[1]] for topic in shown_topics
+                ]
 
-                coherence_scores.append(round(cm.get_coherence(), 5))
+            NMF_stability = {}
+            for i in range(0, len(topic_nums) - 1):
+                jaccard_sims = []
+                for _, topic1 in enumerate(NMF_topics[topic_nums[i]]):
+                    sims = []
+                    for _, topic2 in enumerate(NMF_topics[topic_nums[i + 1]]):
+                        sims.append(self.jaccard_similarity(topic1, topic2))
 
-            scores = list(zip(topic_nums, coherence_scores))
-            chosen_num_topics = sorted(scores, key=itemgetter(1), reverse=True)[0][0]
-        else:
-            chosen_num_topics = num_topics
+                    jaccard_sims.append(sims)
+
+                NMF_stability[topic_nums[i]] = jaccard_sims
+
+            mean_stabilities = [
+                np.array(NMF_stability[i]).mean() for i in topic_nums[:-1]
+            ]
+
+            coherences = [
+                CoherenceModel(
+                    model=NMF_models[i],
+                    texts=texts,
+                    dictionary=dictionary,
+                    coherence="c_v",
+                ).get_coherence()
+                for i in topic_nums[:-1]
+            ]
+
+            coh_sta_diffs = [
+                coherences[i] - mean_stabilities[i]
+                for i in range(len(topic_nums[:-1]))[:-1]
+            ]
+            coh_sta_max = max(coh_sta_diffs)
+            coh_sta_max_idxs = [
+                i for i, j in enumerate(coh_sta_diffs) if j == coh_sta_max
+            ]
+            ideal_topic_num_index = coh_sta_max_idxs[0]
+            chosen_num_topics = topic_nums[ideal_topic_num_index]
 
         model = NMF(n_components=chosen_num_topics, random_state=42)
         model.fit(vectors)
@@ -177,6 +271,8 @@ class ModelResults:
         top_topics = pd.DataFrame(
             np.argmax(component_loadings, axis=1), columns=["top_topic_num"]
         )
+        top_topics["second_topic"] = (-component_loadings).argsort(axis=1)[:, 1]
+        top_topics["third_topic"] = (-component_loadings).argsort(axis=1)[:, 2]
 
         top_topic_loading = pd.DataFrame(
             np.max(component_loadings, axis=1), columns=["top_topic_loading"]
@@ -199,12 +295,26 @@ class ModelResults:
         for topic, comp in enumerate(model.components_):
             word_idx = np.argsort(comp)[::-1][:num_examples]
             topic_words[topic] = [vocab[i] for i in word_idx]
-            # Append examples that have one of the top keywords in topic
-            sample_texts_lst.append(
-                [
+            examples_lst = [
+                x
+                for x in list(
+                    combined_df[combined_df["top_topic_num"] == topic][
+                        text_column_name
+                    ].values
+                )
+                if topic_words[topic][0] in x
+                or topic_words[topic][1] in x
+                or topic_words[topic][2] in x
+                or topic_words[topic][3] in x
+                or topic_words[topic][4] in x
+            ]
+
+            # If not enough examples, check for second topic loading
+            if len(examples_lst) < num_examples:
+                extra_examples_lst_2 = [
                     x
                     for x in list(
-                        combined_df[combined_df["top_topic_num"] == topic][
+                        combined_df[combined_df["second_topic"] == topic][
                             text_column_name
                         ].values
                     )
@@ -213,8 +323,28 @@ class ModelResults:
                     or topic_words[topic][2] in x
                     or topic_words[topic][3] in x
                     or topic_words[topic][4] in x
-                ][:num_examples]
-            )
+                ]
+                examples_lst.extend(extra_examples_lst_2)
+
+            # If not enough examples still, check for third topic loading
+            if len(examples_lst) < num_examples:
+                extra_examples_lst_3 = [
+                    x
+                    for x in list(
+                        combined_df[combined_df["third_topic"] == topic][
+                            text_column_name
+                        ].values
+                    )
+                    if topic_words[topic][0] in x
+                    or topic_words[topic][1] in x
+                    or topic_words[topic][2] in x
+                    or topic_words[topic][3] in x
+                    or topic_words[topic][4] in x
+                ]
+                examples_lst.extend(extra_examples_lst_3)
+
+            # Append examples that have one of the top keywords in topic
+            sample_texts_lst.append(examples_lst[:num_examples])
 
         topic_words_df = pd.DataFrame(
             columns=[
@@ -261,7 +391,7 @@ class ModelResults:
             left_on="top_topic_num",
             right_on="topic_num",
             how="left",
-        ).drop("top_topic_num", axis=1)
+        )[[text_column_name, "topic_num", "top_words_and_phrases"]]
 
         return (
             concated_topics,
@@ -331,7 +461,7 @@ class ModelResults:
                 shap.summary_plot(shap_values, X_test_sample)
                 plt.tight_layout()
                 plt.savefig(f"{path}{class_name}_{filename}.png")
-                
+
             return np.array(lst_all_shap_values)
 
         elif model_type.lower() == "classification":
@@ -348,8 +478,6 @@ class ModelResults:
 
         return shap_values
 
-
-
     def most_common_words_by_group(
         self,
         X,
